from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
from django.db import models


class KeyValueContainer(dict):
    def __init__(self, seq=None, separator="=", **kwargs):
        super().__init__()
        self.sep = separator
        if isinstance(seq, str):
            seq = self._parse_string(seq)
        if seq is not None:
            seq = dict(seq)
            kwargs.update(seq)

        for key, value in kwargs.items():
            self.__setitem__(key, value)

    def __str__(self):
        result = []
        for key, val in self.items():
            result.append(" ".join((key, self.sep, val)))
        return "{0}\n".format("\n".join(result))

    def __setitem__(self, key, item):
        if item is None:
            item = ""
        else:
            item = str(item)
        super().__setitem__(key, item)

    def _parse_string(self, value):
        result = {}
        if not value:
            return result

        for line in value.split("\n"):
            line = line.strip()
            if not line:
                continue
            if self.sep not in line:
                raise ValueError(
                    _("Invalid syntax in line {0}\nExpected: key {1} value").format(
                        repr(line), self.sep
                    )
                )
            key, value = [val.strip() for val in line.split(self.sep, 1)]
            result[key] = value

        return result


class KeyValueField(models.TextField):
    """
    Basically a way to store configuration in DB and have it returned as dictionary.
    Simple key/value store
    data stored as
    key = value
    default separator is "=" but it can be customized

    sample usage

    class MyModel(models.Model):
         options = KeyValueField(separator=":")

    >> my_model.options = "key1 : val1 \n key2 : val2"
    >> my_model.clean_fields()
    >> my_model.options
    {"key1": "val1", "key2": "val2"}
    """

    description = _("Key/Value dictionary field")
    empty_values = (None,)

    def __init__(self, separator="=", *args, **kwargs):
        self.separator = separator
        super().__init__(*args, **kwargs)

    def contribute_to_class(self, cls, name, **kwargs):  # noqa: WPS117
        super().contribute_to_class(cls, name, **kwargs)
        setattr(cls, name, property(fget=self.get_value, fset=self.set_value))

    def set_value(self, obj, value):
        if isinstance(value, str):
            value = self.from_db_value(value)
        elif not isinstance(value, KeyValueContainer):
            value = KeyValueContainer(value)
        obj.__dict__[self.name] = value

    def get_value(self, obj):
        return obj.__dict__[self.name]

    def from_db_value(self, value, *args, **kwargs):
        try:
            return KeyValueContainer(value, separator=self.separator)
        except ValueError as excp:
            raise ValidationError(excp)

    def get_prep_value(self, value):
        if value is None:
            return ""
        return str(value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.separator != "=":
            kwargs["separator"] = self.separator
        return name, path, args, kwargs



from .upload_to_helpers import upload_to, UploadTo
from .cache_helpers import cached_model_property, cached_function
from .model_choices import Choices, Choices
from .key_value_field import KeyValueField



import typing
from django.utils.translation import gettext as _
from collections import OrderedDict


class Choices:
    """
    Offer a cleaner way for django choices field
    Requires python 3.6 or higher
    Improvement over old Choices class:
    1. code completion in IDEs
    2. simpler syntax
    3. supports class inheritance

    Example usage:

    class FruitChoices(model_helpers.Choices2):
        BERRY = {"display": "strawberry", "id": 1, "extra_key": "extra_value"}
        BANANA = 2
        WATER_MELON = 3
        x = "unrelated attr" # this will be ignored because its not in upper case


    FRUITS = FruitChoices()

    # In Django models:
        fruit = models.IntegerField(choices=FRUITS()) # produces [(1, "BERRY"), (2, "BANANA"), (3, "WATER_MELON")]
    """
    _choices = None
    _choices_by_id = None

    def __getattribute__(self, item):
        attr_value = super().__getattribute__(item)
        if item.startswith("_"):
            return attr_value

        if item in self._choices:
            return attr_value["id"]
        return attr_value

    @classmethod
    def _get_all_attrs(cls, class_obj):
        if not class_obj:
            return tuple()

        return cls._get_all_attrs(class_obj.__base__) + tuple(class_obj.__dict__.items())

    def __init__(self, *args):
        if args:
            raise NotImplementedError("Choices class has been updated, please use the new syntax")
        self._choices = {}
        self._choices_by_id = {}
        for attr_name, attr_value in self._get_all_attrs(self.__class__):
            if not attr_name.isupper():
                continue
            attr_name: str
            if not isinstance(attr_value, dict):
                attr_value =  {"id": attr_value}
            if "display" not in attr_value:
                attr_value["display"] = attr_name.replace("_", " ").capitalize()

            attr_value["name"] = attr_name
            choice_id = attr_value["id"]
            self._choices[attr_name] = attr_value
            if choice_id in self._choices_by_id:
                raise ValueError(
                    "Duplicate choice id {choice_id} for {choice_name} and {choice_name2}".format(
                        choice_id=choice_id,
                        choice_name=attr_name,
                        choice_name2=self._choices_by_id[choice_id]["name"],
                    )
                )
            self._choices_by_id[choice_id] = attr_value

    def __call__(self):
        return [
            (choice_value["id"], choice_name)
            for choice_name, choice_value in self._choices.items()
        ]

    def get_choice(self, choice_id) -> dict:
        """
        Return translated display name of certain choice.
        same as model's get_<field_name>_display()
        :param choice_id: choice value as stored in Database
        """
        return self._choices_by_id[choice_id]

    def get_display_name(self, choice_id):
        """
        Return translated display name of certain choice.
        same as model's get_<field_name>_display()
        :param choice_id: choice value as stored in Database
        """
        return self.get_choice(choice_id)["display"]

    def get_choice_name(self, choice_id):
        """
        Returns name of certain choice given its database value.
        :param choice_id: choice value as stored in Database
        """
        return self.get_choice(choice_id)["name"]



import typing
import functools
import json
import types
import inspect
from hashlib import sha256
from django.db import models
from django.core.cache import cache


class CachedFunction:
    def __init__(self, cache_timeout: int, key_parameters: typing.Iterable, key_class_attrs: typing.Iterable):
        """
        cache_timeout: number of seconds to store the cached value. default is determined by Django's `CACHE` settings.
        key_parameters: List of parameter names that need to match for the cache to be considered valid.
        key_class_attrs: List of attributes on the class that needs to match for the cache to be considered valid.
                         Useful only for instance methods or class methods otherwise it doesn't do anything.

        Example1:
            @CachedFunction(key_parameters=["a", "b"])
            def sum(a, b, log_output):
                return a + b

        >> sum(1, b=2, log_output=True)
        3  # sum function is called
        >> sum(1, b=2, log_output=False)  # log_output is not in key_parameters list so it won't be considered
        3  # sum function is NOT called. Cached version is returned instead.
        >> sum(2, b=2)
        4  # sum function is called.

        Example2:
            class A:
                b = 7
                @CachedFunction(key_parameters=["a"], key_class_attrs=["b"])
                def sum(self, a, log_output):
                    return a + self.b

        cache would remain valid so long as `self.b` and `a` variables are unchanged.
        """
        self.cache_timeout = cache_timeout
        self.key_parameters = key_parameters
        self.key_class_attrs = key_class_attrs

    def __call__(self, original_func):
        @functools.wraps(original_func)
        def wrapper(*args, **kwargs):
            cache_key = self.get_cache_key(original_func, *args, **kwargs)
            returned_value = cache.get(cache_key)
            if returned_value is None:
                returned_value = original_func(*args, **kwargs)
                self.cache_value(cache_key, returned_value)
            return returned_value
        return wrapper

    def cache_value(self, cache_key: str, value) -> None:
        """
        Set the cache value of that property
        """
        # Save that key in the cache
        if self.cache_timeout is None:
            cache.set(cache_key, value)
        else:
            cache.set(cache_key, value, self.cache_timeout)

    def get_cache_key(self, function: types.FunctionType, *args, **kwargs) -> str:
        func_arguments: dict = inspect.getcallargs(function, *args, **kwargs)
        if self.key_parameters:
            key_parameters = set(self.key_parameters)
        else:
            key_parameters = set(func_arguments.keys())

        args = {
            param: repr(param_value)
            for param, param_value in func_arguments.items()
            if param in key_parameters
        }

        if "self" in func_arguments or "cls" in func_arguments:
            assert self.key_class_attrs is not None, (
                "Your method is part of a class so, you need to specify which class attributes affect the output. "
                "If output is not affected by class attributes, consider using @staticmethod decorator"
            )
            try:
                class_obj = func_arguments["self"]
            except KeyError:
                class_obj = func_arguments["cls"]
            for class_attr in self.key_class_attrs:
                args["class.{0}".format(class_attr)] = repr(getattr(class_obj, class_attr))

            args.pop("cls", None)
            args.pop("self", None)

        func_signature = {
            "name": function.__qualname__,
            "module": function.__module__,
            "args": args,
        }
        return sha256(json.dumps(func_signature).encode("ascii")).hexdigest()


def cached_function(
    original_func=None, *, cache_timeout=None, key_parameters=None, key_class_attrs=None,
):
    cached_func_instance = CachedFunction(cache_timeout, key_parameters, key_class_attrs)
    if original_func:
        return cached_func_instance(original_func)
    return cached_func_instance


def cached_model_property(  # noqa: WPS212
    model_method=None, readonly=True, cache_timeout=None
):
    """
    cached_model_property is a decorator for model functions that takes no arguments
    The function is converted into a property that support caching out of the box

    :param readonly: set readonly parameter False to make the property writeable
    :type readonly: bool
    :param cache_timeout: number of seconds before cache expires
    :type cache_timeout: int

    Sample usage:

    class Team(models.Model):

        @cached_model_property
        def points(self):
            # Do complex DB queries
            return result

        @cached_model_property(readonly=False)
        def editable_points(self):
            # get result
            return result

    Now try
    team = Team.objects.first()
    team.points  <-- complex DB queries will happen, result will be returned
    team.points  <-- this time result is returned from cache (points function is not called at all!
    del team.points <-- points value has been removed from cache
    team.points  <-- complex DB queries will happen, result will be returned

    set readonly parameter False to make the property writeable
    team.editable_points = 88
    in this case the assigned value will replace the value stored in the cache
    team.editable_points
    returns 88
    """

    def func(original_func):
        cached_func_instance = CachedFunction(cache_timeout, key_class_attrs=["pk"], key_parameters=None)

        def get_x(obj: "models.Model"):
            return cached_func_instance(original_func)(obj)

        def del_x(obj: "models.Model") -> None:
            """
            Remove that property from the cache
            """
            cache_key = cached_func_instance.get_cache_key(original_func, obj)
            # Remove that key from the cache
            cache.delete(cache_key)

        def set_x(obj: "models.Model", value) -> None:
            """
            Set the cache value of that property
            """
            cache_key = cached_func_instance.get_cache_key(original_func, obj)
            cached_func_instance.cache_value(cache_key, value)

        if readonly:
            return property(fget=get_x, fdel=del_x)
        return property(fget=get_x, fset=set_x, fdel=del_x)

    # model_method is passed when using @cached_model_property
    if model_method:
        return func(model_method)
    # model_method is not passed when using @cached_model_property(readonly=True) or even @cached_model_property()
    return func



import types
from os import path as fs_path
from time import strftime
from django.utils.deconstruct import deconstructible
from django.utils.text import slugify
from django.conf import settings


UPLOAD_TO_OPTIONS = types.MappingProxyType(
    {
        "black_listed_extensions": [
            "php",
            "html",
            "htm",
            "js",
            "vbs",
            "py",
            "pyc",
            "asp",
            "aspx",
            "pl",
        ],
        "max_filename_length": 40,
        "file_name_template": "{model_name}/%Y/{filename}.{extension}",
    }
)


@deconstructible
class UploadTo:
    """
    An instance of this class is passed as "upload_to" parameter for any FileField or ImageField
    It ensures file name is less than "max_filename_length" char also slugify the filename and finally provide simple
    protection against uploading some harmful files like (php or python files)

    File is saved in a folder called <model_name>/<current_year>/file_name.ext
    example: User/2015/profile_pic.jpg
    """

    def __init__(self, **kwargs):
        """
        :param kwargs: You can override any of the default options by passing it as keyword argument to this function
        :return:
        """
        self.options = UPLOAD_TO_OPTIONS.copy()
        if hasattr(settings, "UPLOAD_TO_OPTIONS"):
            self.options.update(settings.UPLOAD_TO_OPTIONS)
        self.options.update(kwargs)

    def __call__(self, instance, full_filename):
        """
        :param instance: model instance which the file is uploaded for
        :param full_filename: filename including its path
        :return: string
        """
        file_info = self.get_file_info(full_filename)
        self.validate_file_info(file_info)
        return self.generate_file_name(instance, file_info)

    @staticmethod
    def get_file_info(full_filename):
        filename = fs_path.basename(full_filename).lower()
        filename, file_ext = filename.rsplit(".", 1)
        return {
            "filename": filename,
            "extension": file_ext,
            "full_filename": full_filename,
        }

    def validate_file_info(self, file_info):
        file_ext = file_info["extension"]
        if file_ext in self.options["black_listed_extensions"]:
            raise ValueError("File extension '{0}' is not allowed".format(file_ext))

    def generate_file_name(self, instance, file_info):
        model_name = instance.__class__.__name__
        filename = file_info["filename"]
        max_len = self.options["max_filename_length"]
        file_info["filename"] = slugify(filename)[:max_len]

        return strftime(self.options["file_name_template"]).format(
            model_name=model_name,
            instance=instance,
            **file_info,
        )


# Shortcut for UploadTo class
def upload_to(instance, full_filename):
    upload_to_obj = UploadTo()
    return upload_to_obj(instance, full_filename)
